fn separateVecP2_fn cur_Pos grid_arr targetVec Cross_Corners = 
(
	local p1 = cur_Pos + [targetVec.x,0]
	local p2 = cur_Pos + [0,targetVec.y]
	
	local result = if Cross_Corners then
	(
		(grid_arr[p1.x][p1.y] != #Wall and grid_arr[p2.x][p2.y] != #Wall)
	)
	else
	(
		(grid_arr[p1.x][p1.y] != #Wall or grid_arr[p2.x][p2.y] != #Wall)
	)
	
	return result
)

fn FindFreeCell_fn cur_Pos grid_arr diagonal Cross_Corners =
(
	local dir_arr = #(
		[0,-1],\ 
		[0,1],\ 
		[1,0],\ 
		[-1,0],\ 
		[1,1],\ 
		[1,-1],\ 
		[-1,-1],\ 
		[-1,1]
	)
	
	local newArr = #()
	local numIter = if diagonal then 8 else 4
	local currentW = grid_arr[cur_Pos.x][cur_Pos.y]
	
	for i = 1 to numIter do
	(		
		checkPos = cur_Pos + dir_arr[i]		
		if i < 5 then 
		(
			if grid_arr[checkPos.x][checkPos.y] == (-1) then
			(
				local weight = currentW + 1
				grid_arr[checkPos.x][checkPos.y] = weight
				append newArr checkPos
			)
		)
		else
		(
			if grid_arr[checkPos.x][checkPos.y] == (-1) and (separateVecP2_fn cur_Pos grid_arr dir_arr[i] Cross_Corners) then
			(
				local weight = currentW + sqrt(2)
				grid_arr[checkPos.x][checkPos.y] = weight
				append newArr checkPos
			)
		)
	)
	return newArr
)

fn compareFN v1 v2 =
(
	local d=v1[1]-v2[1]
	case of	
	(
		(d<0.): -1
		(d>0.): 1
		default: 0
	)
)

fn find_path cur_Pos grid_arr diagonal Cross_Corners =
(
	local dir_arr = #(
		[0,-1],\ 
		[0,1],\ 
		[1,0],\ 
		[-1,0],\ 
		[1,1],\ 
		[1,-1],\ 
		[-1,-1],\ 
		[-1,1]
	)
	
	local arr = #()
	local numIter = if diagonal then 8 else 4
	for i = 1 to numIter do
	(	
			checkPos = cur_Pos + dir_arr[i]		
			if i < 5 then 
			(
				if grid_arr[checkPos.x][checkPos.y] != #wall then
				(
					append arr #(grid_arr[checkPos.x][checkPos.y], checkPos)
				)
			)
			else
			(
				if grid_arr[checkPos.x][checkPos.y] != #wall and (separateVecP2_fn cur_Pos grid_arr dir_arr[i] Cross_Corners) then
				(
					append arr #(grid_arr[checkPos.x][checkPos.y], checkPos)
				)
			)
	)

	qsort arr compareFN
	
	return arr[1][2]
)

fn create_shape_fn sPos pathArr rectArr fPos sz =
(
	local sp = splineshape()
	sp.render_displayRenderMesh = true
	sp.render_rectangular = true
	sp.render_width = (sz / 5.0)
	sp.render_length = (sz / 10.0)
	sp.wirecolor = black
	sp.material = standard diffusecolor:(color 255 0 50)
	addnewspline sp
	
	for i in pathArr do
	(
		addknot sp 1 #corner #curve rectArr[i.x][i.y][1].pos
	)
	sp.pos.z = sz * 1.2
	updateshape sp
	
)

fn Root_fn sz w l w_type scl ph freq lvl ofs_x ofs_y w_prctg sp fp diagonal Cross_Corners show_w create_sp = 
(
	delete objects
	
	local wave_arr = #()
	local path_arr = #()
	local start_pos = sp
	local finish_pos = fp
	local findEnd = false
	local free_cell_color = (color 30 30 30)
	local wall_color = (color 100 100 100)
	local wall_material = standard diffusecolor:wall_color
	local start_color = (color 255 55 0)
	local finish_color = (color 0 100 255)
	local wave_color = (color 150 255 0)
	local path_color = (color 0 255 0)

	local grid_arr = for i = 1 to w collect 
					(
						for j = 1 to l collect (-1);
					)
					
	local rect_arr = for i = 1 to w collect 
					(
						if (mod i 5) == 0 then max views redraw;
						for j = 1 to l collect
						(
							b = box pos:([(i - (w / 2.0) - 0.5), (j  - (l / 2.0) - 0.5), 0] * sz) width:(sz / 1.05) length:(sz / 1.05) height:(sz / 9.0) wirecolor:free_cell_color
							b.material = standard diffusecolor:free_cell_color

							t = if show_w then
								(
									text size:(sz / 4.0) pos:([(i - (w / 2.0) - 0.5),((j  - (l / 2.0) - 0.5) - 0.08),-0.1] * sz) text:"0.0" render_displayRenderMesh:false wirecolor:wave_color
								)
								else
								(
									undefined
								)
							#(b,t)
						)
					)
	local iter = 1
	
	grid_arr[start_pos.x][start_pos.y] = 0
	rect_arr[start_pos.x][start_pos.y][1].material.diffusecolor = start_color
	rect_arr[finish_pos.x][finish_pos.y][1].material.diffusecolor = finish_color

	for y = 1 to l do
	(
		if (mod y 5) == 0 then max views redraw;
		for x = 1 to w do
		(
			if y == 1 or y == l then
			(
				grid_arr[x][y] = #wall
				rect_arr[x][y][1].height = (sz / 2.0)
				rect_arr[x][y][1].material = wall_material
			)
			else
			(	
				local newWall = case w_type of
								(
									1:(
										((noise4 ([(x + ofs_x), (y + ofs_y), 0] / scl) ph) + 0.5) >= lvl
									)
									2:(
										((turbulence ([(x + ofs_x), (y + ofs_y), 0] / scl) freq) + 0.5) >= lvl
									)
									3:(
										(random 0.0 100.0) < w_prctg
									)
								)
				if newWall then 
				(
					if [x,y] != start_pos and [x,y] != finish_pos then
					(
						grid_arr[x][y] = #wall
						rect_arr[x][y][1].material = wall_material
						rect_arr[x][y][1].height = (sz / 2.0)
					)
				)
				
				grid_arr[1][y] = #wall
				grid_arr[w][y] = #wall
				rect_arr[1][y][1].material = wall_material
				rect_arr[1][y][1].height = (sz / 2.0)
				rect_arr[w][y][1].material = wall_material
				rect_arr[w][y][1].height = (sz / 2.0)
			)
			
		)
	)
			
	if show_w then
	(
		rect_arr[start_pos.x][start_pos.y][2].wirecolor = black
		rect_arr[finish_pos.x][finish_pos.y][2].wirecolor = black
	)
	
	wave_arr[1] = start_pos

	while wave_arr.count != 0 do
	(
		local new_arr = #()
		
		for j in wave_arr do
		(

			local new_cell_arr = FindFreeCell_fn j grid_arr diagonal Cross_Corners
			join new_arr new_cell_arr
			if j == (finish_pos) then 
			(
				findEnd = true
				
			)
		)
		
		wave_arr = new_arr
		
		if not findEnd then
		(
			animate on
			(
				for j in wave_arr do
				(
					if show_w then
					(
						if j != finish_pos then
						(
							rect_arr[j.x][j.y][2].wirecolor = (wave_color / ((mod iter 2) + 1))
						)
						rect_arr[j.x][j.y][2].text = (((grid_arr[j.x][j.y] * 10.0) as integer) / 10.) as string
					)
					if j != finish_pos then
					(
						at time 0 (
							rect_arr[j.x][j.y][1].height = (sz / 9.0)
							rect_arr[j.x][j.y][1].material.diffusecolor = free_cell_color
							if show_w then rect_arr[j.x][j.y][2].pos.z = (-0.1)
						)
						at time iter (
							rect_arr[j.x][j.y][1].height = (sz / 9.0)
							rect_arr[j.x][j.y][1].material.diffusecolor = free_cell_color
							if show_w then rect_arr[j.x][j.y][2].pos.z = (-0.1)
						)
						at time (iter + 3) (
							rect_arr[j.x][j.y][1].height = sz
							rect_arr[j.x][j.y][1].material.diffusecolor = (wave_color / ((mod iter 2) + 1))
							if show_w then rect_arr[j.x][j.y][2].pos.z = (sz * 1.1)
						)
						at time (iter + 15) (
							rect_arr[j.x][j.y][1].height = (sz / 9.0)
							rect_arr[j.x][j.y][1].material.diffusecolor = free_cell_color
							if show_w then rect_arr[j.x][j.y][2].pos.z = (sz / 8.0)
						)
					)
					else
					(
						at time 0 (
							rect_arr[finish_pos.x][finish_pos.y][1].height = (sz / 9.0)
							if show_w then rect_arr[j.x][j.y][2].pos.z = (-0.1)
						)
						at time iter (
							rect_arr[finish_pos.x][finish_pos.y][1].height = (sz / 9.0)
							if show_w then rect_arr[j.x][j.y][2].pos.z = (-0.1)
						)
						at time (iter + 3) (
							rect_arr[finish_pos.x][finish_pos.y][1].height = sz
							if show_w then rect_arr[j.x][j.y][2].pos.z = (sz * 1.1)
						)
						at time (iter + 15) (
							rect_arr[finish_pos.x][finish_pos.y][1].height = (sz / 9.0)
							if show_w then rect_arr[j.x][j.y][2].pos.z = (sz / 8.0)
						)
					)
				)
			)
		)
		
		iter += 1
		if wave_arr.count == 0 or keyboard.EscPressed or findEnd then exit()
	)
	
	animate on
	(
		at time 3 (
			rect_arr[start_pos.x][start_pos.y][1].height = sz
			if show_w then rect_arr[start_pos.x][start_pos.y][2].pos.z = (sz * 1.1)
		)

		at time 10 (
			rect_arr[start_pos.x][start_pos.y][1].height = (sz / 9.0)
			if show_w then rect_arr[start_pos.x][start_pos.y][2].pos.z = (sz / 8.0)
		)
	)

	append path_arr finish_pos

	if findEnd then
	(
		iter = 0 
		local p = finish_pos
		while p != start_pos do
		(
			p = find_path p grid_arr diagonal Cross_Corners
			if finish_pos == start_pos or keyboard.EscPressed then exit()
			append path_arr p
		)

		animate on
		(
			for i = path_arr.count to 1 by -1 do
			(
				at time (path_arr.count + iter) (
					rect_arr[path_arr[i].x][path_arr[i].y][1].height = (sz / 9.0)
					rect_arr[path_arr[i].x][path_arr[i].y][1].material.diffusecolor = case i of
																					(
																						1:(finish_color)
																						(path_arr.count):(start_color)
																						default:free_cell_color
																					)
				)
				at time (path_arr.count + iter + 5) (
					rect_arr[path_arr[i].x][path_arr[i].y][1].height = (sz / 2.0)
					rect_arr[path_arr[i].x][path_arr[i].y][1].material.diffusecolor = case i of
																					(
																						1:(finish_color)
																						(path_arr.count):(start_color)
																						default:path_color
																					)
				)
	
				at time (path_arr.count + iter + 10) (
					rect_arr[path_arr[i].x][path_arr[i].y][1].height = (sz / 9.0)
					rect_arr[path_arr[i].x][path_arr[i].y][1].material.diffusecolor = case i of
																					(
																						1:(finish_color)
																						(path_arr.count):(start_color)
																						default:path_color
																					)
				)
				if keyboard.EscPressed then exit()
				iter += 1
			) 
		)
		
		if create_sp then
		(
			create_shape_fn start_pos path_arr rect_arr finish_pos sz
		)
	)
)

rollout PathFindingVisualizer "Path Finding Visualizer" width:200 height:596
(
	checkbox chk1 "Allow Diagonal" pos:[16,424] width:96 height:16
	GroupBox grp1 "Cells" pos:[8,8] width:184 height:96
	label lbl1 "Width:" pos:[16,56] width:72 height:16
	label lbl2 "Length:" pos:[16,80] width:72 height:16
	spinner width_spn "" pos:[136,56] width:48 height:16 range:[10,10000,28] type:#integer
	spinner length_spn "" pos:[136,80] width:48 height:16 range:[10,10000,28] type:#integer
	checkbox chk2 "Don't Cross Corners" pos:[16,448] width:120 height:16
	GroupBox grp3 "Position of the start and finish" pos:[8,320] width:184 height:72
	label lbl3 "Start [x,y]:" pos:[16,344] width:72 height:16
	spinner start_p_x "" pos:[80,344] width:48 height:16 range:[2,10000,2] type:#integer
	spinner start_p_y "" pos:[136,344] width:48 height:16 range:[2,10000,2] type:#integer
	label lbl6 "Finish [x,y]:" pos:[16,368] width:72 height:16
	spinner finish_p_x "" pos:[80,368] width:48 height:16 range:[2,10000,27] type:#integer
	spinner finish_p_y "" pos:[136,368] width:48 height:16 range:[2,10000,27] type:#integer
	GroupBox grp4 "Other:" pos:[8,400] width:184 height:120
	button btn1 "[ GO ]" pos:[8,528] width:184 height:40
	GroupBox grp2 "Walls" pos:[8,112] width:184 height:200
	label lbl7 "Type:" pos:[16,139] width:48 height:16
	spinner prctg_w_spn "" pos:[136,288] width:48 height:16 range:[0,100,30]
	dropdownList ddl1 "" pos:[72,136] width:112 height:21 items:#("Noise", "Turbulence", "Random")
	label lbl8 "Percentage of walls:" pos:[16,288] width:104 height:16
	label lbl9 "Scale:" pos:[16,168] width:48 height:16
	spinner scale_spn "" pos:[136,168] width:48 height:16 range:[0,10000,3]
	label lbl10 "Offset [x,y]:" pos:[16,264] width:64 height:16
	spinner ofst_x_spn "" pos:[80,264] width:48 height:16 range:[-10000,10000,0]
	spinner ofst_y_spn "" pos:[136,264] width:48 height:16 range:[-10000,10000,0]
	label lbl14 "Phase:" pos:[16,192] width:48 height:16
	spinner phase_spn "" pos:[136,192] width:48 height:16 range:[0,10000,1]
	label lbl15 "Frequency:" pos:[16,216] width:56 height:16
	spinner freq_spn "" pos:[136,216] width:48 height:16 range:[0,10000,5]
	checkbox chk3 "Show weight" pos:[16,472] width:120 height:16
	checkbox chk4 "Create Spline from Path" pos:[16,496] width:160 height:16
	label lbl47 "Size:" pos:[16,32] width:72 height:16
	spinner cellSize_spn "" pos:[136,32] width:48 height:16 range:[0,10000,5]
	label lbl59 "Level:" pos:[16,240] width:56 height:16
	spinner lvl_spn "" pos:[136,240] width:48 height:16 range:[0.0,1,0.6] scale:0.001
	label lbl_author "Vladislav Bodyul" pos:[8,574]
	hyperlink hp1 "bodyulcg.com" pos:[126,574] address:"http://bodyulcg.com" color:(blue + green)
	
	on PathFindingVisualizer open do
	(
		freq_spn.enabled = false
		prctg_w_spn.enabled = false
	)
	
	on width_spn changed val do
	(
		if start_p_x.value > (val - 1) then start_p_x.value = (val - 1)
		if finish_p_x.value > (val - 1) then finish_p_x.value = (val - 1) 
	)
	
	on length_spn changed val do
	(
		if start_p_y.value > (val - 1) then start_p_y.value = (val - 1)
		if finish_p_y.value > (val - 1) then finish_p_y.value = (val - 1) 
	)
	
	on ddl1 selected sel do
	(
		case sel of
		(
			1:(
				scale_spn.enabled = true
				phase_spn.enabled = true
				freq_spn.enabled = false
				ofst_x_spn.enabled = ofst_y_spn.enabled = true
				prctg_w_spn.enabled = false
			)
			2:(
				scale_spn.enabled = true
				phase_spn.enabled = false
				freq_spn.enabled = true
				ofst_x_spn.enabled = ofst_y_spn.enabled = true
				prctg_w_spn.enabled = false
			)
			3:(
				scale_spn.enabled = false
				phase_spn.enabled = false
				freq_spn.enabled = false
				ofst_x_spn.enabled = ofst_y_spn.enabled = false
				prctg_w_spn.enabled = true
			)
		)
	)
	
	on start_p_x changed val do
	(
		if width_spn.value < (val + 1) then width_spn.value = (val + 1)
	)
	
	on start_p_y changed val do
	(
		if length_spn.value < (val + 1) then length_spn.value = (val + 1)
	)
	
	on finish_p_x changed val do
	(
		if width_spn.value < (val + 1) then width_spn.value = (val + 1)
	)
	
	on finish_p_y changed val do
	(
		if length_spn.value < (val + 1) then length_spn.value = (val + 1)
	)
	
	on btn1 pressed do
	(
		Root_fn\ 
			cellSize_spn.value\ 
			width_spn.value\ 
			length_spn.value\ 
			ddl1.selection\ 
			scale_spn.value\ 
			phase_spn.value\ 
			freq_spn.value\ 
			lvl_spn.value\ 
			ofst_x_spn.value\ 
			ofst_y_spn.value\ 
			prctg_w_spn.value\ 
			[start_p_x.value, start_p_y.value]\ 
			[finish_p_x.value, finish_p_y.value]\ 
			chk1.checked\ 
			chk2.checked\ 
			chk3.checked\ 
			chk4.checked
	)
)
CreateDialog PathFindingVisualizer